import math
import weakref

import carla

from crcarla.visualization.visualization_base import VisualizationBase


class IMUSensor(VisualizationBase):
    """Manages IMU sensor attached to ego vehicle."""

    def __init__(self, parent_actor: carla.Vehicle):
        """
        Initialization of IMU sensor.

        :param parent_actor: Parent CARLA actor.
        """
        super().__init__()  # must have for VisualizationBase inheriting __init__()

        self.sensor = None
        self._parent = parent_actor
        self.accelerometer = (0.0, 0.0, 0.0)
        self.gyroscope = (0.0, 0.0, 0.0)
        self.compass = 0.0
        world = self._parent.get_world()
        bp = world.get_blueprint_library().find("sensor.other.imu")
        self.sensor = world.spawn_actor(bp, carla.Transform(), attach_to=self._parent)
        # We need to pass the lambda a weak reference to self to avoid circular
        # reference.
        weak_self = weakref.ref(self)
        self.sensor.listen(lambda sensor_data: IMUSensor._imu_callback(weak_self, sensor_data))

    @staticmethod
    def _imu_callback(weak_self: weakref, sensor_data: carla.IMUMeasurement):
        """
        Call back function to extract IMU data.

        :param weak_self: Weak self-reference.
        :param sensor_data: CARLA IMU measurement.
        """
        self = weak_self()
        if not self:
            return
        limits = (-99.9, 99.9)
        self.accelerometer = (
            max(limits[0], min(limits[1], sensor_data.accelerometer.x)),
            max(limits[0], min(limits[1], sensor_data.accelerometer.y)),
            max(limits[0], min(limits[1], sensor_data.accelerometer.z)),
        )
        self.gyroscope = (
            max(limits[0], min(limits[1], math.degrees(sensor_data.gyroscope.x))),
            max(limits[0], min(limits[1], math.degrees(sensor_data.gyroscope.y))),
            max(limits[0], min(limits[1], math.degrees(sensor_data.gyroscope.z))),
        )
        self.compass = math.degrees(sensor_data.compass)

    def destroy(self):
        super().destroy()
        self.sensor.stop()
        self.sensor.destroy()
